#!/usr/bin/env python

# Copyright 2015 Lina Sieverling

# This file is part of TelomereHunter.

# TelomereHunter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter.  If not, see <http://www.gnu.org/licenses/>.



import os
from PyPDF2 import PdfFileReader, PdfFileMerger


###############################################################################
### merge all PDF plots produced as TelomereHunter output into one PDF file ###
###############################################################################

def mergeTelomereHunterPDFs(pid, outdir):

    # all possible pdf file names in correct order for merged pdf
    chromosomes = [ str(i) for i in range(1,22+1) ] + ["X","Y"]
    possible_file_names = [pid + '_telomere_content',pid + '_summary', pid + '_hist_telomere_repeats_per_intratelomeric_read', pid + '_gc_content'] + [ pid + "_" + i for i in chromosomes] 
    possible_pdf_names = [ i + '.pdf' for i in possible_file_names ]


    # check which of the possible pdf files exist and sort
    files_dir = os.path.join(outdir, pid, "plots")
    pdf_files = [f for f in os.listdir(files_dir) if f.endswith("pdf")]
    pdf_files_ordered =  [pdf for pdf in possible_pdf_names if pdf in pdf_files]


    # merge files
    merger = PdfFileMerger()

    for filename in pdf_files_ordered:
        merger.append(PdfFileReader(os.path.join(files_dir, filename), "rb"))

    merger.write(os.path.join(outdir, pid, pid + "_merged_plots.pdf"))





#!/usr/bin/python

# Copyright 2015 Lina Sieverling

# This file is part of TelomereHunter.

# TelomereHunter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter.  If not, see <http://www.gnu.org/licenses/>.


############################################################################################
### Returns the repeat threshold from the input BAM file (treshold = read length*6/100). ###
### If the read lengths of the first 100 reads differ, it returns "n".                   ###
############################################################################################

import os
import sys
import pysam


def get_repeat_threshold(bam_file):

      # open input bam_file for reading
      bamfile = pysam.Samfile( bam_file, "rb" )


      ### check if read lengths of the first 100 non-supplementary or secondary alignments are the same
      cntr=0
      different_read_lengths=False

      for read in bamfile.fetch(until_eof=True):   

          if read.is_secondary:        #skips all secondary alignments
              continue

          if read.flag >= 2048:        # skip supplementary alignments
              continue
            
          cntr+=1
          
          if cntr == 1:
              read_length = len(read.seq)
          else:
              if read_length != len(read.seq):
                  different_read_lengths=True
                  repeat_threshold = "n"
                  print "Read lengths in sample differ: repeat threshold will be set individually for each read."
                  break
              
          if cntr == 100:    
              break


      #if they are the same, set threshold
      if different_read_lengths==False:
          repeat_threshold = int(round(float(read_length)*6/100))
          print "Read length is " + str(read_length) + ". Repeat threshold is set to " + str(repeat_threshold) + "."


      return repeat_threshold
    
    


#!/usr/bin/python

# Copyright 2015 Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter.

# TelomereHunter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter.  If not, see <http://www.gnu.org/licenses/>.



import os
import sys
import re
import pysam

#################################################################################
### filters telomere reads from an input BAM file                             ###
### counts the total number of reads mapped to each chromosome band           ###
### determines the gc content distribution of the reads in the input BAM file ###
#################################################################################

def filter_telomere_reads(bam_file, band_file, out_dir, pid, sample, repeat_threshold, mapq_threshold, repeats, consecutive_flag, remove_duplicates):

	################################################
	### get patterns and make regular expression ###
	################################################

	patterns_regex_forward = ""
	patterns_regex_reverse = ""

	for repeat in repeats:
		patterns_regex_forward += repeat + "|"
		patterns_regex_reverse += getReverseComplement(repeat) + "|"

	# remove last '|' from regex
	patterns_regex_forward=patterns_regex_forward[:-1]
	patterns_regex_reverse=patterns_regex_reverse[:-1]


	#########################
	### open file handles ###
	#########################

	# open input bam_file for reading
	bamfile = pysam.Samfile( bam_file, "rb" )

	# open filtered file for writing
	filtered_file_path = out_dir + "/" + pid + "_filtered.bam"
	filtered_file = pysam.Samfile(filtered_file_path, "wb", template=bamfile)



	############################################################################
	### check if repeat threshold needs to be set individually for each read ###
	############################################################################
	
	if repeat_threshold == 'n':
		slow_version=True
	else:
		slow_version=False


	############################
	### make chromosome list ###
	############################

	# autodetect chromosome prefix in BAM file
	references = bamfile.references    # are sorted in the order of the tids

	if references[0][0:3] == "chr":
		bam_chr_prefix = "chr"
	else:
		bam_chr_prefix = ''

	# Generate list of chromosome names.
	chromosome_list = [ str(i) for i in range(1,22+1) ] + ["X","Y"]
	chromosome_list_with_prefix = [bam_chr_prefix + chr for chr in chromosome_list]


	################################################
	### make band, spectrum and gc content lists ###
	################################################

	# Read list of chromosome bands for each chromosome, strip chromosome prefixes.
	bands_list = { chr:{ "band_name":[], "end":[] } for chr in chromosome_list + ["unmapped"] }

	# make spectrum list which will be written to spectrum file
	spectrum_list = { chr:{} for chr in chromosome_list }


	for line in open( band_file, "r" ):

		try:
			line = line.rstrip().split()
			
	#		start = line[1]
			end = line[2]
			band_name = line[3]

			chrom_name = ""

			if line[0][:3] == "chr":
				chrom_name = line[0][3:]
			else:
				chrom_name = line[0]

			bands_list[chrom_name]["band_name"] += [band_name]
			bands_list[chrom_name]["end"] += [int(end)]
			
			spectrum_list[chrom_name][band_name]={"reads":0}

			
		except:

			print( "Invalid line in banding file: \'"+" ".join( line )+"\'" )


	# add unmapped to spectrum list
	spectrum_list["unmapped"] = {"unmapped": {"reads":0 } }

	bands_list["unmapped"]["band_name"] += ["unmapped"]
	bands_list["unmapped"]["end"] += [0]


	# make GC content list
	gc_content_list = { gc_content:0 for gc_content in range(0,100+1) } 



	#############################
	### loop through BAM file ###
	#############################

	# remember last chromosome and it's spectrum list
	lastChromosome = ''
	chromosomeLsEnd = None
	chromosomeLsBand = None
	spectrumTemp = None
	spectrumTemp2 = None
	spectrumUnmapped = spectrum_list["unmapped"]["unmapped"]
	sequence = ""
	chr_offset = len(bam_chr_prefix)
	#initialize band index and chromosome band last looked at
	i = 0
	band = ''


	if slow_version == False:
		for read in bamfile.fetch(until_eof=True):   

			if read.is_secondary:		#skips all secondary alignments (only needed for RNA analysis!)
				continue
			  
			if remove_duplicates == True and read.is_duplicate:		#if duplicate flag is set: skip all reads that are flagged as optical or PCR duplicate
				continue
			      
			if read.flag >= 2048:						# skip supplementary alignments
				continue
			  
			sequence = read.seq

			try:
				read_length = len(sequence)
			except TypeError:                     # skip if there is no sequence for read in BAM file (e.g. * in the sequence field)
				continue
			
			
			# get gc content and add to list (if the fraction of Ns <= 0.2)
			n_count = sequence.count('N')
			
			if float(n_count)/float(read_length) <= 0.2:
				gc_content = int(round(float(sequence.count('C') + sequence.count('G')) / float(read_length - n_count) * 100))
				gc_content_list[gc_content] += 1
		

			# get reference
			tid = read.tid
			ref_name = ''
			if tid != -1:
				ref_name = references[tid]

			#get chromosome and band
			if read.is_unmapped or ref_name not in chromosome_list_with_prefix or read.mapq < mapq_threshold :  # all reads mapped to decoy sequences or with a low mapping quality are defined as unmapped
				spectrumTemp2 = spectrumUnmapped

			else:
				chromosome = ref_name[chr_offset:]

				# for new chromosome
				if chromosome != lastChromosome:
					chromosomeLsEnd = bands_list[chromosome]["end"]
					chromosomeLsBand = bands_list[chromosome]["band_name"]
					lastChromosome = chromosome
					i = 0                              # band index
					band = chromosomeLsBand[i]
					spectrumTemp = spectrum_list[chromosome][band]
				

				read_start_pos = read.pos
				
				# check if read is in new band
				while read_start_pos > chromosomeLsEnd[i] and i <= len(chromosomeLsEnd) :
					i += 1
					band = chromosomeLsBand[i]
					spectrumTemp = spectrum_list[chromosome][band]
				

				spectrumTemp2 = spectrumTemp


			# add +1 to reads mapped to this band
			spectrumTemp2["reads"] += 1
			
			# check if read has the specified amount of patterns, else skip
			if consecutive_flag == False and len(re.findall(patterns_regex_forward, sequence)) < repeat_threshold and len(re.findall(patterns_regex_reverse, sequence)) < repeat_threshold:
				continue
			elif consecutive_flag == True and not re.search("(" + patterns_regex_forward + "){" + str(repeat_threshold) + "}", sequence) and not re.search("(" + patterns_regex_reverse + "){" + str(repeat_threshold) + "}", sequence):
				continue


			# write read to filtered file
			filtered_file.write(read)
			
			
			
			
	else:	# set repeat threshold individually for each read
		for read in bamfile.fetch(until_eof=True):   

			if read.is_secondary:		#skips all secondary alignments (only needed for RNA analysis!)
				continue
			  
			if remove_duplicates == True and read.is_duplicate:		#if duplicate flag is set: skip all reads that are flagged as optical or PCR duplicate
				continue
			      
			if read.flag >= 2048:						# skip supplementary alignments 
				continue
			  
			sequence = read.seq
			
			try:
				read_length = len(sequence)
			except TypeError:                     # skip if there is no sequence for read in BAM file (e.g. * in the sequence field)
				continue
			
			# get gc content and add to list (if the fraction of Ns <= 0.2)
			n_count = sequence.count('N')
			
			if float(n_count)/float(read_length) <= 0.2:
				gc_content = int(round(float(sequence.count('C') + sequence.count('G')) / float(read_length - n_count) * 100))
				gc_content_list[gc_content] += 1
				
				
			# get reference
			tid = read.tid
			ref_name = ''
			if tid != -1:
				ref_name = references[tid]


			#get chromosome and band
			if read.is_unmapped or ref_name not in chromosome_list_with_prefix or read.mapq < mapq_threshold :  # all reads mapped to decoy sequences or with a low mapping quality are defined as unmapped
				spectrumTemp2 = spectrumUnmapped

			else:
				chromosome = ref_name[chr_offset:]

				# for new chromosome
				if chromosome != lastChromosome:
					chromosomeLsEnd = bands_list[chromosome]["end"]
					chromosomeLsBand = bands_list[chromosome]["band_name"]
					lastChromosome = chromosome
					i = 0                              # band index
					band = chromosomeLsBand[i]
					spectrumTemp = spectrum_list[chromosome][band]
				

				read_start_pos = read.pos
				
				# check if read is in new band
				while read_start_pos > chromosomeLsEnd[i] and i <= len(chromosomeLsEnd) :
					i += 1
					band = chromosomeLsBand[i]
					spectrumTemp = spectrum_list[chromosome][band]
				

				spectrumTemp2 = spectrumTemp


			# add +1 to reads mapped to this band
			spectrumTemp2["reads"] += 1
			
			
			repeat_threshold = int(round(float(read_length)*6/100))
			
			# check if read has the specified amount of patterns, else skip
			if consecutive_flag == False and len(re.findall(patterns_regex_forward, sequence)) < repeat_threshold and len(re.findall(patterns_regex_reverse, sequence)) < repeat_threshold:
				continue
			elif consecutive_flag == True and not re.search("(" + patterns_regex_forward + "){" + str(repeat_threshold) + "}", sequence) and not re.search("(" + patterns_regex_reverse + "){" + str(repeat_threshold) + "}", sequence):
				continue


			# write read to filtered file
			filtered_file.write(read)



	#############################
	### write read count file ###
	#############################

	# open spectrum file for writing
	readcount_file = open(out_dir + "/" + pid + "_readcount.tsv", "w")

	# header
	readcount_file.write( "chr\tband\treads\n")

	#write line for each chromosome band
	for chromosome in chromosome_list + ["unmapped"]:

		for band in bands_list[chromosome]["band_name"]:
		  
			readcount_file.write( "%s\t%s\t%i\n" % (chromosome, band, spectrum_list[chromosome][band]["reads"]))


	#############################
	### write gc content file ###
	#############################

	# open gc content file for writing
	gc_content_file = open(out_dir + "/" + pid + "_" + sample + "_gc_content.tsv", "w")

	# header
	gc_content_file.write( "gc_content_percent\tread_count\n")

	#write line for each gc content
	for gc_content in range(0,100+1) :  
	  
		gc_content_file.write( "%i\t%i\n" % (gc_content, gc_content_list[gc_content]))



	##########################
	### close file handles ###
	##########################

	bamfile.close()
	filtered_file.close()
	readcount_file.close()
	gc_content_file.close()

	############################
	### index filtered file  ###  
	############################

	pysam.index(filtered_file_path)



	##################################
	### sort filtered file by name ###  
	##################################

	os.system("samtools sort -n " + filtered_file_path + " " + out_dir + "/" + pid + "_filtered_name_sorted")



# get the reverse complement of a DNA Sequence
def getReverseComplement(sequence):
  
	sequence_temp = sequence.replace("A", "1")
  	sequence_temp = sequence_temp.replace("C", "2")
  	sequence_temp = sequence_temp.replace("G", "3")
  	sequence_temp = sequence_temp.replace("T", "4")
  	
  	sequence_temp2 = sequence_temp.replace("1", "T")
  	sequence_temp2 = sequence_temp2.replace("2", "G")
  	sequence_temp2 = sequence_temp2.replace("3", "C")
  	sequence_temp2 = sequence_temp2.replace("4", "A")
  	
  	sequence_reverse_complement = sequence_temp2[::-1]  # reverses a string
  	
  	return sequence_reverse_complement




#!/usr/bin/python

# Copyright 2015 Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter.

# TelomereHunter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import re
import pysam
import subprocess
from telomerehunter.filter_telomere_reads import getReverseComplement

###############################################################################################################################
### sorts the reads from an input BAM file (needs to be name sorted!) containing telomere reads into 4 different fractions: ###
### intratelomeric, junction-spanning, subtelomeric, intrachromosomal. Additionally, a table containing the number of       ###
### telomere reads and the pattern occurrences per chromosome band is written.                                              ###
###############################################################################################################################

#def sort_telomere_reads(input_dir, band_file, out_dir, pid, mapq_threshold, t_type, c_type, g_type, j_type, n_type):
def sort_telomere_reads(input_dir, band_file, out_dir, pid, mapq_threshold, repeats):

	#####################
	### get patterns  ###
	#####################

	patterns = []

	for repeat in repeats:
		patterns.append(repeat)
		patterns.append(getReverseComplement(repeat))


	# open input bam_file for reading
	bamfile = pysam.Samfile( input_dir + "/" + pid + "_filtered_name_sorted.bam" , "rb" )


	# open filtered files for writing
	intratelomeric_file = pysam.Samfile(out_dir + "/" + pid + "_filtered_intratelomeric.bam", "wb", template=bamfile)
	junctionspanning_file = pysam.Samfile(out_dir + "/" + pid + "_filtered_junctionspanning.bam", "wb", template=bamfile)
	subtelomeric_file = pysam.Samfile(out_dir + "/" + pid + "_filtered_subtelomeric.bam", "wb", template=bamfile)
	intrachromosomal_file = pysam.Samfile(out_dir + "/" + pid + "_filtered_intrachromosomal.bam", "wb", template=bamfile)


	############################
	### make chromosome list ###
	############################

	# get references
	references = bamfile.references 

	# autodetect chromosome prefix in BAM file
	if references[0][0:3] == "chr":
		bam_chr_prefix = "chr"
	else:
		bam_chr_prefix = ''

	# Generate list of chromosome names.
	chromosome_list = [ str(i) for i in range(1,22+1) ] + ["X","Y"]
	chromosome_list_with_prefix = [bam_chr_prefix + chr for chr in chromosome_list]




	####################################
	### make band and spectrum lists ###
	####################################

	# Read list of chromosome bands for each chromosome, strip chromosome prefixes.
	bands_list = { chr:{ "band_name":[], "end":[] } for chr in chromosome_list + ["unmapped"] }

	# make spectrum list which will be written to spectrum file
	spectrum_list = { chr:{} for chr in chromosome_list }

	# add junction_spanning fields for each chromosome to spectrum list
	for chr in chromosome_list:
		spectrum_list[chr]["junction1"] = {pattern:0 for pattern in patterns}
		spectrum_list[chr]["junction1"]["other"]=0.0
		spectrum_list[chr]["junction1"]["reads_with_pattern"]=0
		spectrum_list[chr]["junction2"] = {pattern:0 for pattern in patterns}
		spectrum_list[chr]["junction2"]["reads_with_pattern"]=0
		spectrum_list[chr]["junction2"]["other"]=0.0


	for line in open( band_file, "r" ):

		try:
			line = line.rstrip().split()
			
	#		start = line[1]
			end = line[2]
			band_name = line[3]

			chrom_name = ""

			if line[0][:3] == "chr":
				chrom_name = line[0][3:]
			else:
				chrom_name = line[0]

			bands_list[chrom_name]["band_name"] += [band_name]
			bands_list[chrom_name]["end"] += [int(end)]
			
			spectrum_list[chrom_name][band_name] = {pattern:0 for pattern in patterns}
			spectrum_list[chrom_name][band_name]["other"]=0.0
			spectrum_list[chrom_name][band_name]["reads_with_pattern"]=0
			

		except:

			print( "Invalid line in banding file: \'"+" ".join( line )+"\'" )


	# add unmapped to spectrum list
	spectrum_list["unmapped"] = {"unmapped": {pattern:0 for pattern in patterns} }
	spectrum_list["unmapped"]["unmapped"]["other"]=0.0
	spectrum_list["unmapped"]["unmapped"]["reads_with_pattern"]=0

	bands_list["unmapped"]["band_name"] += ["unmapped"]
	bands_list["unmapped"]["end"] += [0]

	#flag to break out of double while loop
	break_flag = False


	#################################################################
	### go through name sorted BAM file containing telomere reads ###
	### => always look at mate pairs (if possible)                ###
	#################################################################

	while True:			

		try:
			read1 = bamfile.next()
		except: 
			break
		      
		try:
			read2 = bamfile.next()
		except: 
			sort_reads_without_mate (read1, references, bands_list, chromosome_list, mapq_threshold, spectrum_list, patterns,
						intratelomeric_file,
						subtelomeric_file,
						intrachromosomal_file )
			break

		
		
		### reads without mates
		read1_first_flag = True # read 1 is first read

		while read1.qname != read2.qname:
			
			if read1_first_flag == True:
				
				sort_reads_without_mate (read1, references, bands_list, chromosome_list, mapq_threshold, spectrum_list, patterns,
							intratelomeric_file,
							subtelomeric_file,
							intrachromosomal_file )

				try:
					read1 = bamfile.next()
				except: 
					sort_reads_without_mate (read2, references, bands_list, chromosome_list, mapq_threshold, spectrum_list, patterns,
							intratelomeric_file,
							subtelomeric_file,
							intrachromosomal_file )
					break_flag = True
					break	             # break out of inner while loop
				
				read1_first_flag = False # read2 is now first read

				
			else:
				sort_reads_without_mate (read2, references, bands_list, chromosome_list, mapq_threshold, spectrum_list, patterns,
							intratelomeric_file,
							subtelomeric_file,
							intrachromosomal_file )
				
				try:
					read2 = bamfile.next()
				except: 
					sort_reads_without_mate (read1, references, bands_list, chromosome_list, mapq_threshold, spectrum_list, patterns,
							intratelomeric_file,
							subtelomeric_file,
							intrachromosomal_file )
					break_flag = True
					break                 # break out of inner while loop
					
				read1_first_flag = True

		

		if break_flag == True:			# break out of outer while loop if the end of the input BAM file is reached
			break

		### reads with mates

		read1_chromosome, read1_band, read1_is_unmapped, read1_junctionspanning, read1_junction = read_check (read1, references, bands_list, chromosome_list, mapq_threshold)
		read2_chromosome, read2_band, read2_is_unmapped, read2_junctionspanning, read2_junction = read_check (read2, references, bands_list, chromosome_list, mapq_threshold)

		# INTRATELOMERIC: both mates are unmapped
		if read1_is_unmapped==True and read2_is_unmapped==True:

			sort_read_with_mate(read=read1, fraction_file=intratelomeric_file, chromosome="unmapped", band="unmapped", spectrum_list=spectrum_list, patterns=patterns)
			sort_read_with_mate(read=read2, fraction_file=intratelomeric_file, chromosome="unmapped", band="unmapped", spectrum_list=spectrum_list, patterns=patterns)


		# JUNCTION SPANNING: one mate is unmapped and the other mate is mapped to first or last band of chromosome
		elif read1_is_unmapped==True and read2_junctionspanning==True or read2_is_unmapped==True and read1_junctionspanning==True:	

			if read1_is_unmapped==True:
				chromosome = read2_chromosome
				band = read2_junction
			else:
				chromosome = read1_chromosome
				band = read1_junction

			sort_read_with_mate(read=read1, fraction_file=junctionspanning_file, chromosome=chromosome, band=band, spectrum_list=spectrum_list, patterns=patterns)
			sort_read_with_mate(read=read2, fraction_file=junctionspanning_file, chromosome=chromosome, band=band, spectrum_list=spectrum_list, patterns=patterns)	
			
	

		# SUBTELOMERIC: both mates are mapped to first or last band or chromosome (and have similar mapping positions)
		elif read1_junctionspanning==True and read2_junctionspanning==True:     

			sort_read_with_mate(read=read1, fraction_file=subtelomeric_file, chromosome=read1_chromosome, band=read1_band, spectrum_list=spectrum_list, patterns=patterns)
			sort_read_with_mate(read=read2, fraction_file=subtelomeric_file, chromosome=read2_chromosome, band=read2_band, spectrum_list=spectrum_list, patterns=patterns)	


		
		# SUBTELOMERIC/INTRACHROMOSOMAL: one read is subtelomeric, the other is intra-chromosomal => count and sort individually
		elif read1_junctionspanning==True or read2_junctionspanning==True:
		  
			if read1_junctionspanning==True:
				read1_file = subtelomeric_file
				read2_file = intrachromosomal_file
			else:
				read2_file = subtelomeric_file
				read1_file = intrachromosomal_file

			sort_read_with_mate(read=read1, fraction_file=read1_file, chromosome=read1_chromosome, band=read1_band, spectrum_list=spectrum_list, patterns=patterns)
			sort_read_with_mate(read=read2, fraction_file=read2_file, chromosome=read2_chromosome, band=read2_band, spectrum_list=spectrum_list, patterns=patterns)	

		
		
		# INTRACHROMOSOMAL: one read is intra-chromosomal, the other is unmapped => count both to position of mapped read
		elif read1_is_unmapped==True and read2_junctionspanning==False or read2_is_unmapped==True and read1_junctionspanning==False:
			
			if read1_is_unmapped==True:
				chromosome = read2_chromosome
				band = read2_band
			else:
				chromosome = read1_chromosome
				band = read1_band

			sort_read_with_mate(read=read1, fraction_file=intrachromosomal_file, chromosome=chromosome, band=band, spectrum_list=spectrum_list, patterns=patterns)
			sort_read_with_mate(read=read2, fraction_file=intrachromosomal_file, chromosome=chromosome, band=band, spectrum_list=spectrum_list, patterns=patterns)	

		

		# INTRACHROMOSOMAL: both mapped
		else:

			sort_read_with_mate(read=read1, fraction_file=intrachromosomal_file, chromosome=read1_chromosome, band=read1_band, spectrum_list=spectrum_list, patterns=patterns)
			sort_read_with_mate(read=read2, fraction_file=intrachromosomal_file, chromosome=read2_chromosome, band=read2_band, spectrum_list=spectrum_list, patterns=patterns)	




	###########################
	### write spectrum file ###
	###########################

	# open spectrum file for writing
	spectrum_file = open(out_dir + "/" + pid + ".spectrum", "w")

	# header
	spectrum_file.write( "chr\tband\treads_with_pattern")

	for pattern in patterns:
		spectrum_file.write( "\t" + pattern )
		
	spectrum_file.write( "\tother\n" )

	#write line for each chromosome band
	for chromosome in chromosome_list:

		for band in ["junction1"] + bands_list[chromosome]["band_name"] + ["junction2"]:
  
			spectrum_file.write( "%s\t%s\t%i" % (chromosome, band, spectrum_list[chromosome][band]["reads_with_pattern"]))

			for pattern in patterns:
				spectrum_file.write("\t" + str(spectrum_list[chromosome][band][pattern]))

			spectrum_file.write("\t" + str(int(round(spectrum_list[chromosome][band]["other"]))) + "\n" )


	spectrum_file.write( "unmapped\tunmapped\t%i" % (spectrum_list["unmapped"]["unmapped"]["reads_with_pattern"]))

	for pattern in patterns:
		spectrum_file.write("\t" + str(spectrum_list["unmapped"]["unmapped"][pattern]))

	spectrum_file.write("\t" + str(int(round(spectrum_list["unmapped"]["unmapped"]["other"]))) + "\n" )


	##########################
	### close file handles ###
	##########################

	bamfile.close()
	intratelomeric_file.close()
	junctionspanning_file.close()
	subtelomeric_file.close()
	intrachromosomal_file.close()
	spectrum_file.close()



# function for checking reads without mate and sorting
def read_check (read, references, bands_list, chromosome_list, mapq_threshold):
  
	# get reference
	read_tid = read.tid
	read_ref = ''
	if read_tid != -1:
		read_ref = references[read_tid]
		read_ref = read_ref.replace("chr", "")
	
	# get position
	read_pos = read.pos
		
	# get mapping quality
	read_mapq = read.mapq

	
	read_junctionspanning = False
	read_junction = ""
	
	# check if read is considered unmapped
	if read.is_unmapped or read_ref not in chromosome_list or read.mapq < mapq_threshold:
		read_is_unmapped=True
		chromosome = "unmapped"
		band = "unmapped"
	else:
		read_is_unmapped=False
		
		chromosome = read_ref
		
		i=0
		while read_pos > bands_list[chromosome]["end"][i] and i<(len(bands_list[chromosome]["end"])-1):
			i += 1
		else:
			band = bands_list[chromosome]["band_name"][i]
			
			
			# check if read is mapped to first or last chromosome band
			if i==0:
				read_junctionspanning = True
				read_junction = "junction1"
			elif i==(len(bands_list[chromosome]["end"])-1):
				read_junctionspanning = True
				read_junction = "junction2"

	return ( chromosome, band, read_is_unmapped, read_junctionspanning, read_junction )


# function for sorting reads without a mate into correct fraction and add counts to spectrum list
def sort_reads_without_mate (read, references, bands_list, chromosome_list, mapq_threshold, spectrum_list, patterns,
			     intratelomeric_file,
			     subtelomeric_file,
			     intrachromosomal_file ):

	# do read check
	chromosome, band, read_is_unmapped, read_junctionspanning, read_junction = read_check (read, references, bands_list, chromosome_list, mapq_threshold)

	# write read to correct fraction
	if read_is_unmapped:
		intratelomeric_file.write(read)
	elif read_junctionspanning:
		subtelomeric_file.write(read)
	else:
		intrachromosomal_file.write(read)

	# add read counts to spectrum list
	spectrum_list[chromosome][band]["reads_with_pattern"]+=1

	read_total_pattern_count = 0
	
	for pattern in patterns:
		spectrum_list[chromosome][band][pattern] += read.seq.count( pattern )
		read_total_pattern_count += read.seq.count( pattern )


	spectrum_list[chromosome][band]["other"] += float(len(read.seq))/6 - read_total_pattern_count


# function for writing reads with mates to fraction and adding counts to spectrum list
def sort_read_with_mate (read, fraction_file, chromosome, band, spectrum_list, patterns):

	fraction_file.write(read)
	
	spectrum_list[chromosome][band]["reads_with_pattern"]+=1

	read_total_pattern_count = 0

	#count occurrence of pattern types	
	for pattern in patterns:
		spectrum_list[chromosome][band][pattern] += read.seq.count( pattern )
		read_total_pattern_count += read.seq.count( pattern )

	spectrum_list[chromosome][band]["other"] += float(len(read.seq))/6 - read_total_pattern_count




#!/usr/bin/python

# Copyright 2015 Lina Sieverling

# This file is part of TelomereHunter.

# TelomereHunter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import re
import pysam
from telomerehunter.filter_telomere_reads import getReverseComplement

###############################################################################################
### get the distribution of telomeric repeats per intratelomeric read in the input BAM file ###
###############################################################################################

#def repeat_frequency_intratelomeric(input_path, out_dir, pid, t_type, c_type, g_type, j_type):
def repeat_frequency_intratelomeric(input_path, out_dir, pid, repeats):
	################################################
	### get patterns and make regular expression ###
	################################################

	patterns_regex_forward = ""
	patterns_regex_reverse = ""
		
	for repeat in repeats:
		patterns_regex_forward += repeat + "|"
		patterns_regex_reverse += getReverseComplement(repeat) + "|"

	# remove last '|' from regex
	patterns_regex_forward=patterns_regex_forward[:-1]
	patterns_regex_reverse=patterns_regex_reverse[:-1]

		
	#########################
	### open file handles ###
	#########################

	# open input bam_file for reading
	bamfile = pysam.Samfile( input_path + "/" + pid + "_filtered_intratelomeric.bam", "rb" )


	##################################
	### initialize frequency table ###
	##################################

	frequency_table = { repeats:0 for repeats in range(0,16+1) }

	######################################
	### loop through filtered BAM file ###
	######################################

	for read in bamfile.fetch(until_eof=True):   

		sequence = read.seq

		number_repeats_forward = len(re.findall(patterns_regex_forward, sequence))
		number_repeats_reverse = len(re.findall(patterns_regex_reverse, sequence))
		
		if number_repeats_forward > number_repeats_reverse:
			number_repeats = number_repeats_forward
		else:
			number_repeats = number_repeats_reverse

		try:
		      frequency_table[number_repeats] += 1
		except:
		      frequency_table[number_repeats] = 1       # if key does not exist: add to frequency_table


	##################################
	### write frequency table file ###
	##################################

	# open frequency table file for writing
	frequency_table_file = open(out_dir + "/" + pid + "_repeat_frequency_per_intratelomeric_read.tsv", "w")

	# header
	frequency_table_file.write( "number_repeats\tcount\n")

	#write line for each chromosome band
	for frequency in frequency_table:

		frequency_table_file.write( "%i\t%i\n" % (frequency, frequency_table[frequency]))


	##########################
	### close file handles ###
	##########################

	bamfile.close()
	frequency_table_file.close()





#setup file

from setuptools import setup

setup(name='telomerehunter',
      version='1.0.1',
      description='Estimation of Telomere Content from WGS Data',
      long_description='TelomereHunter extracts, sorts and analyses telomeric reads from WGS Data. It is designed to take BAM files from a tumor and/or a control sample as input. The tool was developed at the German Cancer Research Center (DKFZ).',
      keywords='telomere content read NGS WGS tumor control',
      url='https://pypi.python.org/pypi/telomerehunter/',
      author='Lars Feuerbach <l.feuerbach@dkfz-heidelberg.de>, Philip Ginsbach, Lina Sieverling <l.sieverling@dkfz-heidelberg.de>',
      author_email='l.sieverling@dkfz-heidelberg.de',
      license='GPL',
      packages=['telomerehunter'],
      install_requires=['pysam', 'PyPDF2'],
      include_package_data = True,
      scripts = ['bin/telomerehunter'],
      zip_safe=False)



#!/usr/bin/python

# Copyright 2015 Lina Sieverling, Philip Ginsbach, Lars Feuerbach

# This file is part of TelomereHunter.

# TelomereHunter is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# TelomereHunter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with TelomereHunter.  If not, see <http://www.gnu.org/licenses/>.



import subprocess
import os
import sys
import re
import csv
import pysam


##################################################################
### get the gc content distribution of the reads in a bam file ###
##################################################################

def get_gc_content_distribution(bam_file, out_dir, pid, sample, remove_duplicates):

	# open input bam_file for reading
	bamfile = pysam.Samfile( bam_file, "rb" )

	# make GC content list
	gc_content_list = { gc_content:0 for gc_content in range(0,100+1) }


	for read in bamfile.fetch(until_eof=True):   

		if read.is_secondary:		#skips all secondary alignments (only needed for RNA analysis!)
			continue
		  
		if remove_duplicates == True and read.is_duplicate:		#if duplicate flag is set: skip all reads that are flagged as optical or PCR duplicate
			continue
		      
		if read.flag >= 2048:						# skip supplementary alignments 
			continue
		  
		sequence = read.seq
		read_length = len(read.seq)
		
		c_count = sequence.count('C')
		g_count = sequence.count('G')
		n_count = sequence.count('N')
		

		if float(n_count)/float(read_length) > 0.2:
			continue
		
		gc_content = int(round(float(c_count + g_count) / float(read_length - n_count) * 100))
		
		gc_content_list[gc_content] += 1
		

	#############################
	### write gc content file ###
	#############################

	# open gc content file for writing
	gc_content_file = open(out_dir + "/" + pid + "_" + sample + "_gc_content.tsv", "w")

	# header
	gc_content_file.write( "gc_content_percent\tread_count\n")

	#write line for each gc content
	for gc_content in range(0,100+1) :
	  
		gc_content_file.write( "%i\t%i\n" % (gc_content, gc_content_list[gc_content]))


	##########################
	### close file handles ###
	##########################

	bamfile.close()
	gc_content_file.close()




####################################################################################################
### estimate the telomere content in telomeric reads per million reads with a similar gc content ###
####################################################################################################

def estimate_telomere_content(input_dir, out_dir, pid, sample, gc_lower, gc_upper):
	
	#gc bins used for correction
	gc_bins_correction = range(gc_lower, gc_upper+1)
	
	# read in total gc content counts
	gc_content_file = input_dir + "/" + pid + "_" + sample + "_gc_content.tsv"

	gc_content_list = {}

	with open(gc_content_file,'rb') as tsvin:
	    tsvin = csv.reader(tsvin, delimiter='\t')
	    
	    next(tsvin, None)  # skip the headers

	    for row in tsvin:
		gc_content_list[int(row[0])]=int(row[1])


	# get total number of reads in these bins
	sum_over_threshold=0

	for gc in gc_bins_correction:
		sum_over_threshold+=gc_content_list[gc]


	# get total read count and number of intratelomeric_reads
	read_count_file = input_dir + "/" + pid + '_readcount.tsv'
	total_read_count = subprocess.Popen('less ' + read_count_file + ' | awk \'{if (NR!=1) {print}}\'  | awk \'{ SUM += $3}  END { print SUM }\'' , shell=True, stdout=subprocess.PIPE).stdout.read()
	total_read_count = int(total_read_count.strip())

	filtered_intratelomeric_file = input_dir + "/" + pid + '_filtered_intratelomeric.bam'
	intratel_read_count = subprocess.Popen('samtools view -c ' + filtered_intratelomeric_file, shell=True, stdout=subprocess.PIPE).stdout.read()
	intratel_read_count = int(intratel_read_count.strip())


	#estimate telomere content (intratelomeric reads per million reads with similar gc content)
	tel_content = float(intratel_read_count) / float(sum_over_threshold) *1000000


	##########################
	### write summary file ###
	##########################

	# open summary file for writing
	summary_file = open(out_dir + "/" + pid + "_" + sample + "_summary.tsv", "w")

	# header
	#summary_file.write( "PID\tsample\ttotal_reads\tintratel_reads\tintratel_gc_count_threshold\ttotal_reads_with_tel_gc\ttel_content\n")
	summary_file.write( "PID\tsample\ttotal_reads\tintratel_reads\tgc_bins_for_correction\ttotal_reads_with_tel_gc\ttel_content\n")
	
	# results 
	gc_bins = str(gc_lower) + "-" + str(gc_upper)
	summary_file.write( "%s\t%s\t%i\t%i\t%s\t%i\t%f\n" % (pid, sample, total_read_count, intratel_read_count, gc_bins, sum_over_threshold, tel_content))

	##########################
	### close file handles ###
	##########################

	summary_file.close()
	
